"""xTB interface functions"""
import logging
import shutil
import subprocess
import tempfile
from pathlib import Path

from rdkit import Chem
from rdkit.Chem import rdmolops

from lefqm import constants

XTB_DEFAULT = {
    "gfn": 2,
    "alpb": "water",
    "opt": None,
    "cycles": 100,
}

XTB_FAST = {
    "gfn": 0,
    "alpb": "water",
    "opt": None,
    "cycles": 200,
}

XTB_LAX = {
    "gfn": 2,
    "alpb": "water",
    "opt": "lax",
}


def xtb_optimize_conformation(mol, conformation_index, xtb="xtb", cores=None):
    """Optimize a conformation with XTB

    :param mol: molecule with conformations to optimize
    :type mol: rdkit.Chem.rdchem.Mol
    :param xtb: path/call to xtb
    :type xtb: str
    :param conformation_index: index of the conformation to optimize
    :type conformation_index: int
    :return: optimized conformation and XTB conformer energy in hartrees
    :rtype: (rdkit.Chem.rdchem.Conformer, float)
    """
    if not shutil.which(xtb):
        raise RuntimeError(f"Cannot find {xtb}")

    with tempfile.TemporaryDirectory() as xtb_dir:
        logging.debug("Optimizing conformers in: %s", xtb_dir)
        xtb_dir = Path(xtb_dir)
        charge = rdmolops.GetFormalCharge(mol)
        with open(xtb_dir / ".CHRG", "w", encoding="utf8") as charge_file:
            charge_file.write(str(charge))

        input_file_path = xtb_dir / "input.xyz"
        with open(input_file_path, "w", encoding="utf8") as input_file:
            input_file.write(Chem.MolToXYZBlock(mol, confId=conformation_index))

        command = [xtb, input_file_path.name]
        for option, value in XTB_DEFAULT.items():
            command.append(f"--{option}")
            if value is not None:
                command.append(str(value))

        if cores:
            command.extend(["--parallel", str(cores)])

        xtb_log_path = xtb_dir / "xtb.log"
        logging.debug(" ".join(command))
        try:
            with open(xtb_log_path, "w", encoding="utf8") as log_file:
                subprocess.check_call(
                    command, cwd=xtb_dir, stdout=log_file, stderr=subprocess.STDOUT
                )
        except subprocess.CalledProcessError:
            logging.info(open(xtb_log_path, encoding="utf8").read())
            raise

        with open(xtb_log_path, encoding="utf8") as log_file:
            lines = list(log_file.readlines())
            if "finished run on" not in lines[-16]:
                raise RuntimeError("XTB calculation did not converge")

        with open(xtb_dir / "xtbopt.xyz", encoding="utf8") as optimized_xyz_file:
            xyz_string = optimized_xyz_file.read()

        comment_line = [
            token.strip() for token in xyz_string.split("\n")[1].split(" ") if token.strip()
        ]
        assert "energy:" == comment_line[0], "Energy was not written with molecule"
        energy = float(comment_line[1])
        optimized_mol = Chem.MolFromXYZBlock(xyz_string)

    return optimized_mol.GetConformer(0), energy


def xtb_optimize(mol, xtb="xtb", conformation_indexes=None, cores=None):
    """Optimize the conformations of a molecule with XTB

    :param mol: molecule with conformations to optimize
    :type mol: rdkit.Chem.rdchem.Mol
    :return: optimized molecule and XTB conformer energies in kcal/mol
    :rtype: (rdkit.Chem.rdchem.Mol, list[float])
    """
    if conformation_indexes is None:
        conformation_indexes = range(mol.GetNumConformers())

    optimized_mol = Chem.Mol(mol, True)  # copy without conformers
    optimized_mol.SetProp("_Name", mol.GetProp("_Name"))
    energies = []
    for index, conformation_index in enumerate(conformation_indexes):
        try:
            optimized_conformer, energy = xtb_optimize_conformation(
                mol, conformation_index, xtb=xtb, cores=cores
            )
            energies.append(energy * constants.HARTREE_TO_KCALMOL)
            optimized_mol.AddConformer(optimized_conformer, assignId=True)
        except RuntimeError as error:
            logging.warning(
                "Skipped conformation in molecule %s that encountered error: %s",
                mol.GetProp("_Name"),
                error,
            )
        logging.info("Optimized %s / %s", index + 1, len(conformation_indexes))
    return optimized_mol, energies



"""Gaussian interface functions"""
import logging
import re
import shutil
import subprocess
import tempfile
from pathlib import Path

from rdkit import Chem
from rdkit.Chem import rdmolops

GAUSSIAN_TEMPLATE = """
# B3LYP/cc-pVDZ NMR SCRF=SMD

Shieldings

{charge} 1
{xyz_string}

"""

SHIELDING_PATTERN = re.compile(r"\s*[0-9]+\s*[a-zA-Z]+\s*Isotropic\s*=\s*((\-|\+)?[0-9]+\.[0-9]+)")


def gaussian_read_isotropic_shieldings(log_path):
    """Read isotropic shielding from guassian log file

    :param log_path: path to the guassian log file
    :type log_apth: pathlib.Path
    :return: isotropic shielding constants
    :rtype: list[(int,str,float)]
    """
    isotropic_shielding_constants = []
    with open(log_path, encoding="utf8") as log_file:
        for line in log_file.readlines():
            if "Isotropic" not in line:
                continue

            shielding_match = SHIELDING_PATTERN.match(line)
            if shielding_match is None:
                continue

            value = shielding_match.group(1)
            try:
                value = float(value)
            except ValueError as error:
                logging.warning(error)
                value = None
            isotropic_shielding_constants.append(value)
    return isotropic_shielding_constants


def gaussian_calculate_shieldings(mol, gaussian="g16", run_dir_path=None):
    """Calculate gaussian shieldings for a molecule

    :param mol: molecule to calculate shieldings for
    :type mol: rdkit.Chem.rdchem.Mol
    :param gaussian: path/call to gaussian
    :type gaussian: str
    :param run_dir_path: path to the directory to run in
    :type run_dir_path: pathlib.Path
    :return: shieldings for every atom
    :rtype: list[float]
    """
    if not shutil.which(gaussian):
        raise RuntimeError(f"Cannot find {gaussian}")

    tmp_dir = None
    if run_dir_path is None:
        tmp_dir = tempfile.TemporaryDirectory()
        run_dir_path = Path(tmp_dir.name)
    logging.debug("Calculating shieldings in %s", run_dir_path)

    xyz_string = "\n".join(Chem.MolToXYZBlock(mol).split("\n")[2:]).strip()
    input_string = GAUSSIAN_TEMPLATE.format(
        charge=rdmolops.GetFormalCharge(mol), xyz_string=xyz_string
    )
    input_name = "shielding.com"
    with open(run_dir_path / input_name, "w", encoding="utf8") as input_file:
        input_file.write(input_string)

    args = [gaussian, input_name]
    logging.debug(" ".join(args))
    log_path = run_dir_path / "shielding.log"
    try:
        subprocess.check_call(args, cwd=run_dir_path)
    except subprocess.CalledProcessError:
        logging.info(open(log_path, encoding="utf8").read())
        raise

    shielding_constants = gaussian_read_isotropic_shieldings(log_path)

    if tmp_dir is not None:
        tmp_dir.cleanup()

    return shielding_constants



"""QM utils"""
import configparser
import copy
import io
import itertools
import logging
import math

import numpy as np
from lefshift import utils

from lefqm import constants


def get_fluorine_shieldings(mol, shielding_property=constants.SHIELDING_SD_PROPERTY):
    """Get fluorine shieldings for a molecule

    TODO make shieldings a type

    CF3 and CF2 fluorine shieldings are averaged. The keys for CF3 and CF2
    groups are tuples.

    :param mol: molecule to extract shieldings from
    :type mol: rdkit.Chem.rdchem.Mol
    :return: dict of atom indexes to shieldings
    :rtype: dict[int | tuple[int], float]
    """
    fluorine_atoms = {atom for atom in mol.GetAtoms() if atom.GetSymbol() == "F"}
    flourine_shieldings = {}
    processed_indexes = set()
    # CF3
    for cf3_group in itertools.combinations(fluorine_atoms, 3):
        if not utils.have_common_neighbor(cf3_group):
            continue
        cf3_group_indexes = tuple(sorted([atom.GetIdx() for atom in cf3_group]))
        if any(index in processed_indexes for index in cf3_group_indexes):
            continue
        flourine_shieldings[cf3_group_indexes] = sum(
            atom.GetDoubleProp(shielding_property) for atom in cf3_group
        ) / len(cf3_group)
        processed_indexes.update(cf3_group_indexes)
    # CF2
    for cf2_group in itertools.combinations(fluorine_atoms, 2):
        if not utils.have_common_neighbor(cf2_group):
            continue
        cf2_group_indexes = tuple(sorted([atom.GetIdx() for atom in cf2_group]))
        if any(index in processed_indexes for index in cf2_group_indexes):
            continue
        flourine_shieldings[cf2_group_indexes] = sum(
            atom.GetDoubleProp(shielding_property) for atom in cf2_group
        ) / len(cf2_group)
        processed_indexes.update(cf2_group_indexes)
    not_processed_indexes = [
        atom.GetIdx() for atom in fluorine_atoms if atom.GetIdx() not in processed_indexes
    ]
    # CF
    for atom_index in not_processed_indexes:
        flourine_shieldings[atom_index] = mol.GetAtomWithIdx(atom_index).GetDoubleProp(
            shielding_property
        )
    return flourine_shieldings


def get_boltzmann_weights(energies, temperature=298.15, boltzmann_constant=1.987204259e-3):
    """Calculate boltzmann weights

    :param energies: energies in kcal per Mol
    :type energies: list[float]
    :param temperature: temperature in Kelvin
    :type temperature: float
    :param boltzmann_constant: Boltzmann constant in kcal/(mol x K)
    :type boltzmann_constant: float
    :return: weights
    :rtype: np.ndarray[float]
    """
    if energies is None or len(energies) == 0:
        raise RuntimeError("Passed invalid energies")

    inverse_kt = 1.0 / (boltzmann_constant * temperature)
    energies = np.array(copy.copy(energies))
    energies -= energies.min()
    energies = np.exp(-energies * inverse_kt)
    energy_sum = np.sum(energies)
    return energies / energy_sum


def get_averaged_shielding(shieldings):
    """Get a single averaged shielding for multiple shieldings

    The lowest key will be associated with the new averaged shielding

    :param shieldings: shieldings to average
    :type shieldings: dict[int | tuple[int], float]
    :return: single averaged shielding
    :rtype: dict[int | tuple[int], float]
    """
    average = 0
    lowest_key = None
    lowest_key_value = None
    for key in shieldings:
        average += shieldings[key]
        key_value = key
        if isinstance(key, tuple):
            key_value = sum(key)
        if lowest_key_value is None or lowest_key_value > key_value:
            lowest_key_value = key_value
            lowest_key = key
    average /= len(shieldings)
    return {lowest_key: average}


def get_weighted_average_ensemble_shieldings(ensemble_shieldings, weights):
    """Calculate weighted averages of the ensemble shieldings

    :param ensemble_shieldings: ensemble shieldings to average
    :type ensemble_shieldings: list[dict[int | tuple[int], float]]
    :param weights: weights to average by
    :type weights: list[float]
    :return: shieldings average over the ensemble
    :rtype: dict[int | tuple[int], float]
    """
    normalized_weights = np.array(copy.copy(weights))
    if not math.isclose(np.sum(normalized_weights), 1.0):
        normalized_weights /= np.sum(normalized_weights)
    averaged_shieldings = {}
    for key in ensemble_shieldings[0].keys():
        averaged_shieldings[key] = sum(
            shieldings[key] * weight
            for shieldings, weight in zip(ensemble_shieldings, normalized_weights)
        )
    return averaged_shieldings


def match_shieldings(ensemble_shieldings, ensemble):
    """Match shieldings of an ensemble

    :param ensemble_shieldings: ensemble shieldings to match
    :type ensemble_shieldings: list[dict[int | tuple[int], float]]
    :param ensemble: ensemble of molecule isomers
    :type ensemble; list[rdkit.Chem.rdchem.Mol]
    :return: matched shieldings
    :rtype: list[dict[int | tuple[int], float]]
    """
    # first set of shieldings is used as reference
    reference_shieldings = ensemble_shieldings[0]
    reference_mol = ensemble[0]
    matched_ensemble_shieldings = []
    for shieldings, mol in zip(ensemble_shieldings, ensemble):
        if reference_shieldings.keys() == shieldings.keys():
            matched_ensemble_shieldings.append(shieldings)
            continue
        match_map = utils.map_atom_indexes(mol, reference_mol)
        matched_shieldings = {}
        for key in shieldings.keys():
            if isinstance(key, tuple):
                new_key = []
                for index in key:
                    new_key.append(match_map[index])
                new_key = tuple(new_key)
                matched_shieldings[new_key] = shieldings[key]
            else:
                matched_shieldings[match_map[key]] = shieldings[key]
        matched_ensemble_shieldings.append(matched_shieldings)
    return matched_ensemble_shieldings


def get_ensemble_fluorine_shieldings(
    ensemble,
    water_energy_property=constants.ENERGY_SD_PROPERTY,
    shielding_property=constants.SHIELDING_SD_PROPERTY,
):
    """Get fluorine shieldings for an ensemble of molecules

    :param ensemble: ensemble of molecule isomers
    :type ensemble; list[rdkit.Chem.rdchem.Mol]
    :return: dict of atom indexes to shieldings of the ensemble
    :rtype: dict[int | tuple[int], float]
    """
    ensemble_fluorine_shieldings = [
        get_fluorine_shieldings(mol, shielding_property=shielding_property) for mol in ensemble
    ]
    ensemble_fluorine_shieldings = match_shieldings(ensemble_fluorine_shieldings, ensemble)
    energies = [float(mol.GetProp(water_energy_property)) for mol in ensemble]
    conformer_weights = get_boltzmann_weights(energies)
    weighted_average_shieldings = get_weighted_average_ensemble_shieldings(
        ensemble_fluorine_shieldings, conformer_weights
    )
    return weighted_average_shieldings


def get_lowest_energy_fluorine_shieldings(
    ensemble,
    water_energy_property=constants.ENERGY_SD_PROPERTY,
    shielding_property=constants.SHIELDING_SD_PROPERTY,
):
    """Get fluorine shieldings for an ensemble of molecules

    :param ensemble: ensemble of molecule isomers
    :type ensemble; list[rdkit.Chem.rdchem.Mol]
    :return: dict of atom indexes to shieldings of the ensemble or None
    :rtype: dict[int | tuple[int], float] or None
    """
    energies = [float(mol.GetProp(water_energy_property)) for mol in ensemble]
    lowest_energy = min(energies)
    for energy, mol in zip(energies, ensemble):
        if math.isclose(energy, lowest_energy):
            return get_fluorine_shieldings(mol, shielding_property=shielding_property)

    return None


def get_config(config_file_path=constants.DEFAULT_CONFIG):
    """Read default and user-specified configuration"""
    config = configparser.ConfigParser()
    config.read(constants.DEFAULT_CONFIG)
    if config_file_path != constants.DEFAULT_CONFIG:
        config.read(config_file_path)
    with io.StringIO() as string_stream:
        string_stream.write("Config:\n")
        config.write(string_stream)
        string_stream.seek(0)
        logging.debug(string_stream.read())
    return config



"""Conformator interface functions"""
import logging
import shutil
import subprocess
import tempfile
from pathlib import Path

from rdkit import Chem
from rdkit.Chem import SDMolSupplier


def conformator_generate(mol, conformator="conformator", max_confs=250, run_dir_path=None):
    """Generate conformations for a molecule using the conformator

    Conformator commandline interface:

        Calculates conformations for the molecules in the input file:

        Options:
         -h [ --help ]                  Prints help message
         -v [ --verbosity ] arg (=3)    Set verbosity level
                                        (0 = Quiet, 1 = One-line-summary, 2 = Errors,
                                        3 = Warnings, 4 = Info)
         -i [ --input ] arg             Input file, suffix is required.
         -o [ --output ] arg            Output file, suffix is required.
         -q [ --quality ] arg (=2)      Set quality level
                                        (1 = Fast, 2 = Best)
         -n [ --nOfConfs ] arg (=250)   Set maximum number of conformations to be
                                        generated.
         -f [ --from ] arg (=1)         Position of first entry in the calculation.
         -t [ --to ] arg (=4294967295)  Position of last entry in the calculation.
         --keep3d                       Keep initial 3D coordinates for molecule as
                                        starting point for conformation generation.
         --hydrogens                    Consider hydrogen clashes during conformation
                                        generation.
         --macrocycle_size arg (=10)    Define minimum size of macrocycles (<= 10)
         --rmsd_input                   Calculate the minimum RMSD of the closest
                                        ensemble member to the input structure.
         --rmsd_ensemble                Calculate the minimum RMSD of the closest
                                        ensemble members to each other.

        License:
         --license arg                  To reactivate the executable, please provide a
                                        new license key.

    :param mol: mol to generate conformations for
    :type mol: rdkit.Chem.rdchem.Mol
    :param conformator: path/call to conformator
    :type conformator: str
    :param max_confs: maximum number of conformations to generate
    :type max_confs: int
    :param run_dir_path: path to the directory to run in
    :type run_dir_path: pathlib.Path
    :return: mol with conformations
    :rtype: rdkit.Chem.rdchem.Mol
    """
    if not shutil.which(conformator):
        raise RuntimeError(f"Cannot find {conformator}")

    tmp_dir = None
    if run_dir_path is None:
        tmp_dir = tempfile.TemporaryDirectory()
        run_dir_path = Path(tmp_dir.name)

    logging.debug("Generating conformers in: %s", run_dir_path)

    input_file_path = run_dir_path / "input.smi"
    with open(input_file_path, "w", encoding="utf8") as input_file:
        input_file.write(Chem.MolToSmiles(mol) + " " + mol.GetProp("_Name"))

    output_file_path = run_dir_path / "output.sdf"
    logging.debug("Output file is: %s", output_file_path)

    args = [
        conformator,
        "-i",
        str(input_file_path),
        "-o",
        str(output_file_path),
        "--nOfConfs",
        str(max_confs),
    ]
    logging.debug(" ".join(args))
    shell_output = subprocess.check_output(args, stderr=subprocess.STDOUT)
    if shell_output:
        logging.debug(shell_output)
    if (
        bytes("INPUT", encoding="ascii") not in shell_output
        and bytes("NONE", encoding="ascii") not in shell_output
        and bytes("PURE", encoding="ascii") not in shell_output
    ):
        logging.warning("Detected potential undefined stereo")
    if not output_file_path.exists():
        raise RuntimeError("Did not generate conformations")

    mols = list(SDMolSupplier(str(output_file_path), removeHs=False))
    mol = mols[0]
    for conformer_mol in mols[1:]:
        mol.AddConformer(conformer_mol.GetConformer(0), assignId=True)
    logging.info("Generated %s conformations", mol.GetNumConformers())

    if tmp_dir is not None:
        tmp_dir.cleanup()
    return mol



"""Process QM ensemble fluorine shielding calculations"""
import logging
from pathlib import Path

from lefshift.utils import map_atom_indexes
import pandas as pd
from rdkit import Chem
from rdkit.Chem import SDMolSupplier, SmilesParserParams

from lefqm import constants, utils


def add_ensembles_subparser(subparsers):
    """Add ensembles arguments as a subparser"""
    ensemble_parser = subparsers.add_parser("ensembles")
    ensemble_parser.add_argument(
        "input", help="Input directory of ensemble SDFs with QM properties", type=Path
    )
    ensemble_parser.add_argument("output", help="Output shielding CSV", type=Path)
    ensemble_parser.add_argument(
        "--lowest",
        action="store_true",
        help="Use only the lowest energy conformation instead of the ensemble",
    )
    ensemble_parser.add_argument(
        "--energy-property",
        type=str,
        help="Property that contains the energy of the conformer in water",
        default=constants.ENERGY_SD_PROPERTY,
    )
    ensemble_parser.add_argument(
        "--shielding-property",
        type=str,
        help="Property that contains the shieldings of the conformer",
        default=constants.SHIELDING_SD_PROPERTY,
    )
    ensemble_parser.add_argument(
        "-v", "--verbose", help="show verbose output", action="store_true"
    )


def make_output(calculated_shieldings):
    """Make output from ensemble shieldings

    :param calculated_shieldings: molecule and shielding pairs to write
    :type calculated_shieldings: list[rdkit.Chem.rdchem.Mol, dict]
    :return: output to write
    :rtype: pd.DataFrame
    """
    lines = []
    params = SmilesParserParams()
    params.removeHs = False
    for mol, shieldings in calculated_shieldings:
        name = mol.GetProp("_Name")
        # Prepare mapping atom indices to the way the molecule will be written
        # as smiles. The input was SDF so the atom order will change.
        smiles = Chem.MolToSmiles(mol)
        canon_mol = Chem.MolFromSmiles(smiles, params)
        atom_index_map = map_atom_indexes(mol, canon_mol)
        for key, shielding in shieldings.items():
            if isinstance(key, tuple):
                if len(key) == 2:
                    label = "CF2"
                elif len(key) == 3:
                    label = "CF3"
                else:
                    raise RuntimeError("Invalid fluorine group")
                # pick the first atom index as a representative
                key = key[0]
            else:
                label = "CF"

            lines.append([name, smiles, label, atom_index_map[key], shielding])
    return pd.DataFrame(
        lines,
        columns=[
            constants.ID_COLUMN,
            constants.SMILES_COLUMN,
            constants.LABEL_COLUMN,
            constants.ATOM_INDEX_COLUMN,
            constants.SHIELDING_COLUMN,
        ],
    )


def ensembles(args):
    """Process QM ensemble fluorine shielding calculations"""
    calculated_shieldings = []
    files = []
    if args.input.is_dir():
        files = sorted(args.input.glob("*.sdf"))
    else:
        files.append(args.input)
    logging.info("Combining ensembles from %s file(s)", len(files))

    for ensemble_path in files:
        ensemble = list(SDMolSupplier(str(ensemble_path), removeHs=False))
        if len(ensemble) == 0:
            raise RuntimeError(f"No molecules in {ensemble_path}")
        ensemble = [
            mol
            for mol in ensemble
            if mol.HasProp(args.energy_property)
            and mol.HasProp("atom.dprop." + args.shielding_property)
        ]
        if len(ensemble) == 0:
            raise RuntimeError(f"No molecules with shieldings in {ensemble_path}")

        if any(not mol.HasProp("_Name") for mol in ensemble):
            logging.warning(
                'Some molecules in "%s" did not have names. This ensemble will be output using the filename: "%s"',
                ensemble_path,
                ensemble_path.stem,
            )
            for mol in ensemble:
                mol.SetProp("_Name", ensemble_path.stem)
        if args.lowest:
            fluorine_shieldings = utils.get_lowest_energy_fluorine_shieldings(ensemble)
        else:
            fluorine_shieldings = utils.get_ensemble_fluorine_shieldings(
                ensemble,
                water_energy_property=args.energy_property,
                shielding_property=args.shielding_property,
            )
        if not fluorine_shieldings:
            raise RuntimeError("No shieldings")
        calculated_shieldings.append((ensemble[0], fluorine_shieldings))

    output = make_output(calculated_shieldings)
    output.to_csv(args.output, index=False)



"""MoKa interface functions"""
import logging
import shutil
import subprocess
import tempfile
from pathlib import Path

from rdkit.Chem import SDMolSupplier, SDWriter


def moka_protonate(mol, moka="blabber_sd", ph=7.6, min_abundance=90):
    """Protonate molecule

    Assign the most abundant tautomer/protomer. MoKa commandline interface:

        usage: /usr/prog/MoKa/4.0.1/blabber_sd.bin [options] <filename>

         <filename> can be a SD (.sd or .sdf) or MOL2 (.mol2) file

        options are:
         -h, --help                     display this help
             --version                  show version info
             --input-type=<sd|mol2|smi> input file type (autodetect)
         -p <pH list>, --pH=<pH list>   report most abundant species
                                        at given pH (7.4)
         -n, --neutralize               report only the neutral species
             --load-model=<database>    use a custom model database for predictions
         -o, --output=FILE              output file name (SD only)
         -m, --minimize                 perform full 3D minimization
         -t <abundance>                 report all species above the threshold
         -e, --equivalent-protomers     generate all equivalent protomers

        <pH list> is a comma separated list of the following:
            * single value
            * range ( min-max )

    :param mol: mol to generate tautomers/protomers for
    :type mol: rdkit.Chem.rdchem.Mol
    :param moka: path/call to moka
    :type moka: str
    :param ph: pH for protonation
    :type ph: float
    :param min_abundance: cutoff in percent for protomer abundance
    :type min_abundance: int
    :return: protonated mol
    :rtype: rdkit.Chem.rdchem.Mol
    """
    if not shutil.which(moka):
        raise RuntimeError(f"Cannot find {moka}")

    with tempfile.TemporaryDirectory() as moka_dir:
        moka_dir = Path(moka_dir)
        logging.debug("Generating tautomers/protomers in: %s", moka_dir)

        input_file_path = moka_dir / "input.sdf"
        writer = SDWriter(str(input_file_path))
        writer.SetForceV3000(True)
        writer.write(mol)
        writer.close()

        output_file_path = moka_dir / "output.sdf"
        logging.debug("Output file is: %s", output_file_path)

        command = [
            moka,
            str(input_file_path),
            "-o",
            str(output_file_path),
            "-p",
            str(ph),
            "-t",
            str(min_abundance),
        ]
        logging.debug(" ".join(command))
        shell_output = subprocess.check_output(command, stderr=subprocess.STDOUT)
        if shell_output:
            logging.debug(shell_output)

        mols = SDMolSupplier(str(output_file_path), removeHs=False)
        if len(mols) == 1 and mols[0].GetProp("ABUNDANCE") == "No species found":
            logging.debug("MoKa found no protomers, returning original")
            return mol

        mols = sorted(
            mols, key=lambda mol: float(mol.GetProp("ABUNDANCE").split("%")[0]), reverse=True
        )
    return mols[0]



"""Generate representative conformations"""
import logging
from pathlib import Path

import numpy as np
import pandas as pd
from lefshift.application_utils import validate_column
from rdkit import Chem
from rdkit.Chem import AllChem, SDWriter, StereoSpecified, rdmolops
from rdkit.Chem.EnumerateStereoisomers import EnumerateStereoisomers, StereoEnumerationOptions
from rdkit.Chem.MolStandardize import rdMolStandardize
from rdkit.Chem.SaltRemover import SaltRemover
from scipy.cluster.hierarchy import fcluster, linkage

from lefqm import constants, utils
from lefqm.commandline_calculation import ConformerGeneration
from lefqm.moka import moka_protonate
from lefqm.xtb import xtb_optimize


def add_conformers_subparser(subparsers):
    """Add conformers arguments as a subparser"""
    conformers_parser = subparsers.add_parser("conformers")
    conformers_parser.add_argument("input", type=Path, help="input SMILES or CSV file")
    conformers_parser.add_argument("output", type=Path, help="output directory for conformer SDFs")
    conformers_parser.add_argument(
        "--ensemble-size", type=int, help="size of the conformer ensemble", default=15
    )
    conformers_parser.add_argument(
        "--id-column",
        help="name of the column containing the ID",
        default=constants.ID_COLUMN,
        type=str,
    )
    conformers_parser.add_argument(
        "--smiles-column",
        help="name of the column containing the molecule SMILES",
        default=constants.SMILES_COLUMN,
        type=str,
    )
    conformers_parser.add_argument("--chunk", type=int, help="which chunk to process")
    conformers_parser.add_argument("--chunk-size", type=int, help="chunk size to use", default=10)
    conformers_parser.add_argument(
        "--cores", type=int, help="Maximum number of cores to use. Defaults to XTB behavior"
    )
    conformers_parser.add_argument(
        "-v", "--verbose", help="show verbose output", action="store_true"
    )
    conformers_parser.add_argument(
        "--config",
        type=Path,
        help="Config file to read from",
        default=constants.DEFAULT_CONFIG,
    )


def get_smiles_chunk(
    smiles_path,
    chunk=None,
    chunk_size=None,
    id_column=constants.ID_COLUMN,
    smiles_column=constants.SMILES_COLUMN,
):
    """Get SMILES chunk to process

    Get all if no chunking arguments were passed

    :param smiles_path: path to SMILES file
    :type smiles_path: pathlib.Path
    :param chunk: number of the chunk
    :type chunk: int
    :param chunk_size: size of the chunk
    :type chunk_size: int
    :return: list of SMILES making up the chunk
    :rtype: list[str]
    """
    start_index = None
    end_index = None
    if chunk is not None:
        start_index = chunk * chunk_size
        end_index = start_index + chunk_size

    smiles_chunk = []
    if smiles_path.suffix == ".smi":
        with open(smiles_path, encoding="utf8") as smiles_file:
            for index, line in enumerate(smiles_file.readlines()):
                if not line:
                    continue
                if end_index is not None and index >= end_index:
                    break
                if start_index is None or index >= start_index:
                    smiles_chunk.append(line.strip())
    elif smiles_path.suffix == ".csv":
        smiles_data = pd.read_csv(smiles_path, dtype={id_column: str, smiles_column: str})
        smiles_data = validate_column(
            smiles_data,
            smiles_column,
            str,
            f'Could not find structure column "{smiles_column}" in input. Specify structure column with name "--smiles-column" option.',
        )
        smiles_data = validate_column(
            smiles_data,
            id_column,
            str,
            f'Could not find ID column "{id_column}" in input. Specify structure column with name "--id-column" option.',
        )
        if start_index is not None and end_index is not None:
            smiles_data = smiles_data[start_index:end_index]
        smiles_chunk = list(smiles_data[smiles_column] + " " + smiles_data[id_column])

    if chunk is not None and len(smiles_chunk) == 0:
        raise RuntimeError(f"Invalid chunk {chunk} has no members. (chunks start with --chunk 0)")

    return smiles_chunk


def normalize(mol):
    """Normalize a mol

    Normalization does the following:
    1. Strip salts
    2. Uncharge
    3. Assign stereo centers

    More than one unassigned stereocenter will cause an error

    :param mol: mol to normalize
    :type mol: rdkit.Chem.rdchem.Mol
    :return: normalized mol
    :rtype: rdkit.Chem.rdchem.Mol
    """
    stripped_mol = SaltRemover().StripMol(mol)
    if stripped_mol.GetNumAtoms() == 0:
        logging.warning(
            'Salt strippper removed entire molecule "%s", keeping original', mol.GetProp("_Name")
        )
    else:
        mol = stripped_mol
    if rdmolops.GetFormalCharge(mol) != 0:
        mol = rdMolStandardize.ChargeParent(mol)
        Chem.SanitizeMol(mol)
        Chem.AssignStereochemistry(mol, cleanIt=True, force=True, flagPossibleStereoCenters=True)

    stereo_specified = [info.specified for info in rdmolops.FindPotentialStereo(mol)]
    unspecified_count = stereo_specified.count(StereoSpecified.Unspecified)
    if unspecified_count > 1 or (unspecified_count == 1 and len(stereo_specified) > 1):
        raise RuntimeError("Cannot process unassigned diastereomers")

    if unspecified_count == 1:
        stereo_options = StereoEnumerationOptions(
            tryEmbedding=True, onlyUnassigned=True, maxIsomers=1, rand=67
        )
        stereo_isomers = list(EnumerateStereoisomers(mol, options=stereo_options))
        assert len(stereo_isomers) >= 1, "RDKit should generate stereo isomers"
        mol = stereo_isomers[0]
    return mol


def prune_conformations(mol, prune_threshold):
    """Prune duplicated conformations

    :param mol: molecule with conformations
    :type mol: rdkit.Chem.rdchem.Mol
    :param prune_threshold: RMSD threshold of molecule to consider duplicates
    :type prune_threshold: float
    """
    nof_conformations = mol.GetNumConformers()
    rms_matrix = AllChem.GetConformerRMSMatrix(mol, prealigned=False)
    full_rms_matrix = np.zeros((nof_conformations, nof_conformations))
    condensed_index = 0
    for i in range(1, nof_conformations):
        for j in range(i):
            full_rms_matrix[i, j] = rms_matrix[condensed_index]
            full_rms_matrix[j, i] = rms_matrix[condensed_index]
            condensed_index += 1

    # prune close conformations
    remove = set()
    for i in range(nof_conformations):
        if i in remove:
            continue

        for j in range(i + 1, nof_conformations):
            if full_rms_matrix[i, j] < prune_threshold:
                remove.add(j)

    for i in remove:
        mol.RemoveConformer(i)
    nof_conformations = mol.GetNumConformers()
    logging.debug("%s left after pruning by %s RMSD", nof_conformations, prune_threshold)


def rmsd_cluster(
    mol, conformation_energies, nof_clusters=15, nof_picks_cluster=1, prune_threshold=0.15
):
    """Cluster conformations by pairwise RMSD

    :param mol: molecule with conformations
    :type mol: rdkit.Chem.rdchem.Mol
    :param conformation_energies: energies of the conformations
    :type conformation_energies: list[float]
    :param nof_clusters: maximum number of clusters to generate
    :type nof_clusters: int
    :param nof_picks_cluster: number of conformations to pick per cluster
    :type nof_picks_cluster: int
    :param prune_threshold: RMSD threshold of molecule to consider duplicates
    :type prune_threshold: float
    :return: indexes of the picked conformations
    :rtype: list[int]
    """
    # copy mol to not modify the input mol
    mol = Chem.Mol(mol)
    assert (
        len(conformation_energies) == mol.GetNumConformers()
    ), "Must have energies for all conformations"

    # prune conformations
    prune_conformations(mol, prune_threshold)
    conformer_ids = [conformer.GetId() for conformer in mol.GetConformers()]
    if mol.GetNumConformers() <= nof_clusters:
        return conformer_ids

    # actual clustering
    rms_matrix = AllChem.GetConformerRMSMatrix(mol, prealigned=True)
    matrix = linkage(rms_matrix, method="ward", metric="euclidean")
    cluster_labels = fcluster(matrix, nof_clusters, criterion="maxclust")

    clusters = {}
    for index, label in enumerate(cluster_labels):
        if label not in clusters:
            clusters[label] = []
        clusters[label].append((index, conformation_energies[index]))

    cluster_picks = []
    for cluster in clusters.values():
        cluster_indexes, _ = zip(*sorted(cluster, key=lambda x: x[1]))
        cluster_picks.extend(cluster_indexes[:nof_picks_cluster])

    logging.debug("Clustered down to %s conformations", len(cluster_picks))
    # map cluster picks back to conformer IDs
    return [conformer_ids[pick] for pick in cluster_picks]


def conformers(args):
    """Generate representative conformations"""
    config = utils.get_config(args.config)
    config["Parameters"]["cores"] = str(args.cores)

    if args.input.suffix not in {".smi", ".csv"}:
        raise RuntimeError("Input may only be a SMILES (.smi) or a CSV (.csv)")

    args.output.mkdir(exist_ok=True, parents=True)
    for smiles in get_smiles_chunk(
        args.input, args.chunk, args.chunk_size, args.id_column, args.smiles_column
    ):
        try:
            mol = Chem.MolFromSmiles(smiles)
            if not mol.HasProp("_Name"):
                logging.warning(
                    "Molecule %s does not have a name. The SMILES will be used.", smiles
                )
                mol.SetProp("_Name", smiles)

            mol = normalize(mol)
            mol = moka_protonate(mol, moka=config["Paths"]["moka"])
            mol = ConformerGeneration(config).run(mol)

            # optimization
            mol, energies = xtb_optimize(mol, xtb=config["Paths"]["xtb"], cores=args.cores)

            # RMSD clustering
            conformation_indexes = rmsd_cluster(
                mol,
                energies,
                nof_clusters=args.ensemble_size,
                nof_picks_cluster=1,
                prune_threshold=float(config["Parameters"]["conf_prune_threshold"]),
            )

            writer = SDWriter(str(args.output / (mol.GetProp("_Name") + ".sdf")))
            writer.SetForceV3000(True)
            conformation_energies = [energies[index] for index in conformation_indexes]
            boltzmann_weights = utils.get_boltzmann_weights(np.array(conformation_energies))
            for conformer_id, energy, weight in zip(
                conformation_indexes, conformation_energies, boltzmann_weights
            ):
                mol.SetProp(constants.ENERGY_SD_PROPERTY, str(energy))
                mol.SetProp(constants.BOLTZMANN_WEIGHT_SD_PROPERTY, str(weight))
                writer.write(mol, confId=conformer_id)
            writer.close()
        except Exception as exception:
            logging.warning("Molecule %s failed with error: %s", smiles, exception)



"""Turbomole interface functions"""
import logging
import shutil
import subprocess
import tempfile
from pathlib import Path

from rdkit import Chem
from rdkit.Chem import rdmolops

TURBOMOLE_CONTROL = """
$atoms
basis=def2-TZVP
$coord file=coord
$symmetry c1
$eht charge={charge}
$dft
functional xcfun set-gga
functional xcfun kt3 1.0
gridsize 3
$rij
$marij
$cosmo
  epsilon=78.39
$nmr dft shielding constants file=shielding
$end
"""


def x2t_convert_mol(mol, x2t="x2t", run_dir_path=None):
    """Convert mol to turbomole coord input

    :param mol: molecule to convert
    :type mol: rdkit.Chem.rdchem.Mol
    :param run_dir_path: path to the directory to run in
    :type run_dir_path: pathlib.Path
    :return: turbomole coord input
    :rtype: str
    """
    if not shutil.which(x2t):
        raise RuntimeError(f"Cannot find {x2t}")

    tmp_dir = None
    if run_dir_path is None:
        tmp_dir = tempfile.TemporaryDirectory()
        run_dir_path = Path(tmp_dir.name)
    logging.debug("Converting mol in %s", run_dir_path)

    xyz_path = run_dir_path / "input.xyz"
    with open(xyz_path, "w", encoding="utf8") as xyz_file:
        xyz_file.write(Chem.MolToXYZBlock(mol))

    command = [x2t, str(xyz_path)]
    logging.debug(" ".join(command))
    return subprocess.check_output(command, cwd=run_dir_path).decode("ascii")


def turbomole_read_isotropic_shieldings(shieldings_path):
    """Read isotropic shieldings from turbomole shielding file

    :param shieldings_path: path to the turbomole shielding file
    :type shieldings_path: pathlib.Path
    :return: isotropic shielding constants
    :rtype: list[float]
    """
    with open(shieldings_path, encoding="utf8") as shieldings_file:
        lines = shieldings_file.readlines()

    shielding_constants = []
    # skip header line
    for line in lines[1:]:
        if "$end" in line:
            break
        if line.strip()[0] == "#":
            continue
        # [NO., TYPE, MULT., ISOTROPIC, ANISOTROPIC, dD/dB-CONTRIBUTION]
        tokens = [token.strip() for token in line.split(" ") if token.strip()]
        atom_name = tokens[1]
        if "*" in atom_name:
            atom_name = atom_name.replace("*", "")
        shielding_constants.append(float(tokens[3]))
    return shielding_constants


def turbomole_calculate_shieldings(
    mol,
    x2t="x2t",
    ridft="ridft",
    mpshift="mpshift",
    cores=None,
    run_dir_path=None,
):
    """Calculate shieldings for a molecule

    :param mol: molecule to calculate shieldings for
    :type mol: rdkit.Chem.rdchem.Mol
    :param x2t: path/call to x2t
    :type x2t: str
    :param ridft: path/call to ridft
    :type ridft: str
    :param mpshift: path/call to mpshift
    :type mpshift: str
    :param cores: maximum number of cores to use
    :type cores: int
    :param run_dir_path: path to the directory to run in
    :type run_dir_path: pathlib.Path
    :return: shieldings for every atom
    :rtype: list[float]
    """
    if not shutil.which(ridft):
        raise RuntimeError(f"Cannot find {ridft}")
    if not shutil.which(mpshift):
        raise RuntimeError(f"Cannot find {mpshift}")

    tmp_dir = None
    if run_dir_path is None:
        tmp_dir = tempfile.TemporaryDirectory()
        run_dir_path = Path(tmp_dir.name)
    logging.debug("Calculating shieldings in %s", run_dir_path)

    coord_output = x2t_convert_mol(mol, x2t=x2t, run_dir_path=run_dir_path)
    coord_path = run_dir_path / "coord"
    with open(coord_path, "w", encoding="utf8") as coord_file:
        coord_file.write(coord_output)

    control_path = run_dir_path / "control"
    with open(control_path, "w", encoding="utf8") as control_file:
        control_file.write(TURBOMOLE_CONTROL.format(charge=rdmolops.GetFormalCharge(mol)))

    ridft_log_path = run_dir_path / "ridft.log"
    with open(ridft_log_path, "w", encoding="utf8") as ridft_log_file:
        ridft_command = [ridft]
        if cores is not None:
            ridft_command.extend(["-nthreads", str(cores)])
        logging.debug(" ".join(ridft_command))
        subprocess.check_call(
            ridft_command,
            cwd=run_dir_path,
            stdout=ridft_log_file,
            stderr=subprocess.STDOUT,
        )

    with open(ridft_log_path, encoding="utf8") as ridft_log_file:
        lines = list(ridft_log_file.readlines())
        if ": all done  ****" not in lines[-6] or "did not converge!" in lines[-15]:
            logging.debug("\n".join(lines))
            raise RuntimeError("ridft calculation did not converge")

    mpshift_log_path = run_dir_path / "mpshift.log"
    with open(mpshift_log_path, "w", encoding="utf8") as mpshift_log_file:
        mpshift_command = [mpshift]
        if cores is not None:
            mpshift_command.extend(["-nthreads", str(cores)])
        logging.debug(" ".join(mpshift_command))
        subprocess.check_call(
            mpshift_command,
            cwd=run_dir_path,
            stdout=mpshift_log_file,
            stderr=subprocess.STDOUT,
        )

    with open(mpshift_log_path, encoding="utf8") as mpshift_log_file:
        lines = list(mpshift_log_file.readlines())
        if ": all done  ****" not in lines[-6]:
            logging.debug("\n".join(lines))
            raise RuntimeError("mpshift calculation did not converge")

    shieldings_path = run_dir_path / "shielding"
    shielding_constants = turbomole_read_isotropic_shieldings(shieldings_path)

    if tmp_dir is not None:
        tmp_dir.cleanup()

    return shielding_constants



"""Shielding generation"""
import logging
from pathlib import Path

from rdkit import Chem
from rdkit.Chem import SDMolSupplier, SDWriter

from lefqm import constants, utils
from lefqm.commandline_calculation import ShieldingCalculation


def add_shieldings_subparser(subparsers):
    """Add shieldings arguments as a subparser"""
    shieldings_parser = subparsers.add_parser("shieldings")
    shieldings_parser.add_argument("input", type=Path, help="Input SD file")
    shieldings_parser.add_argument("output", type=Path, help="Output SD file with shieldings")
    shieldings_parser.add_argument(
        "--cores",
        type=int,
        help="Maximum number of cores to use. Defaults to turbomole behavior",
    )
    shieldings_parser.add_argument(
        "-v", "--verbose", help="show verbose output", action="store_true"
    )
    shieldings_parser.add_argument(
        "--config",
        type=Path,
        help="Config file to read from",
        default=constants.DEFAULT_CONFIG,
    )


def shieldings(args):
    """Shielding generation"""
    config = utils.get_config(args.config)
    config["Parameters"]["cores"] = str(args.cores)

    writer = SDWriter(str(args.output))
    writer.SetForceV3000(True)
    for index, mol in enumerate(SDMolSupplier(str(args.input), removeHs=False)):
        mol_name = mol.GetProp("_Name") if mol.HasProp("_Name") else str(index)
        try:
            logging.info("Processing molecule %s", mol_name)
            shielding_constants = ShieldingCalculation(config).run(mol)
            for atom, shielding in zip(mol.GetAtoms(), shielding_constants):
                atom.SetDoubleProp(constants.SHIELDING_SD_PROPERTY, shielding)
            Chem.CreateAtomDoublePropertyList(mol, constants.SHIELDING_SD_PROPERTY)
            writer.write(mol)
        except Exception as exception:
            logging.warning("Molecule %s failed with error: %s", mol_name, exception)
    logging.info("done")



"""Shift calculation"""
import logging
from pathlib import Path

import pandas as pd
from lefshift.application_utils import validate_column
from sklearn.linear_model import LinearRegression

from lefqm import constants


def add_shifts_subparser(subparsers):
    """Add shifts arguments as a subparser"""
    shifts_parser = subparsers.add_parser("shifts")
    shifts_parser.add_argument(
        "input", help="Input CSV with shieldings to convert to chemical shifts", type=Path
    )
    shifts_parser.add_argument(
        "--calibration", required=True, help="CSV calibration data for the conversion", type=Path
    )
    shifts_parser.add_argument("output", help="OUtput CSV with chemical shifts", type=Path)
    shifts_parser.add_argument(
        "--shielding-column",
        help="name of the column containing the shieldings constants",
        default=constants.SHIELDING_COLUMN,
        type=str,
    )
    shifts_parser.add_argument(
        "--shift-column",
        help="name of the column containing the chemical shift",
        default=constants.SHIFT_COLUMN,
        type=str,
    )
    shifts_parser.add_argument(
        "--id-column",
        help="name of the column containing the ID",
        default=constants.ID_COLUMN,
        type=str,
    )
    shifts_parser.add_argument("-v", "--verbose", help="show verbose output", action="store_true")


def shifts(args):
    """Convert shieldings in input to shifts"""
    data = pd.read_csv(args.input)
    id_column_error = f'Could not find ID column "{args.id_column}" in input. Specify ID column name with "--id-column" option.'
    shielding_column_error = f'Could not find shielding constants column "{args.shielding_column}" in input. Specify shielding column name with "--shielding-column" option.'
    shift_column_error = f'Could not find chemical shift column "{args.shift_column}" in input. Specify shift column name with "--shift-column" option.'
    data = validate_column(data, args.id_column, str, id_column_error)
    data = validate_column(data, args.shielding_column, float, shielding_column_error)

    calibration_data = pd.read_csv(args.calibration)
    calibration_data = validate_column(calibration_data, args.id_column, str, id_column_error)
    calibration_data = validate_column(
        calibration_data, args.shielding_column, float, shielding_column_error
    )
    calibration_data = validate_column(
        calibration_data, args.shift_column, float, shift_column_error
    )
    calibration_data = validate_column(calibration_data, constants.LABEL_COLUMN, str)

    all_shifts = []
    for cf_label in constants.CF_LABELS:
        logging.info('Calibrating conversion of "%s" shieldings', cf_label)
        current_calibration_data = calibration_data[
            calibration_data[constants.LABEL_COLUMN] == cf_label
        ]
        if len(current_calibration_data) == 0:
            logging.warning('No calibration data found for "%s", skipping', cf_label)
            continue

        calibration = LinearRegression().fit(
            current_calibration_data[args.shielding_column].values.reshape(-1, 1),
            current_calibration_data[args.shift_column],
        )

        logging.info('Converting "%s" shieldings', cf_label)
        current_data = data[data[constants.LABEL_COLUMN] == cf_label]
        if len(current_data) == 0:
            logging.warning('No input data found for "%s", skipping', cf_label)
            continue

        chemical_shifts = calibration.predict(
            current_data[args.shielding_column].values.reshape(-1, 1)
        )
        shifts_df = pd.DataFrame(
            chemical_shifts, columns=[args.shift_column], index=current_data.index
        )
        all_shifts.append(shifts_df)

    all_shifts_df = pd.concat(all_shifts)
    data_with_shifts = data.join(all_shifts_df, rsuffix=" Calculated")
    data_with_shifts.to_csv(args.output, index=False)



"""Fluorine chemical shift prediction"""
import argparse
import logging

from lefqm.conformers import add_conformers_subparser, conformers
from lefqm.ensembles import add_ensembles_subparser, ensembles
from lefqm.shieldings import add_shieldings_subparser, shieldings
from lefqm.shifts import add_shifts_subparser, shifts

DESCRIPTION = """
LEFQM calculates fluorine chemical shifts using QM techniques

examples:

Generate conformers from SMILES file or a CSV containing SMILES into a
directory:

lefqm conformers input.smi conformer_ensembles/
lefqm conformers input.csv conformer_ensembles/ --smiles-column 'Name of SMILES column'  # specify which column contains the SMILES

SMILES and CSV files should contain IDs/names. These will be used to name the
output conformer ensembles. If these are not given the conformer ensembles will
be named by the input SMILES.

Calculate NMR shielding constants for an SDF containing molecules/conformers
using Turbomole:

lefqm shieldings conformers.sdf shieldings.sdf

Combine NMR shielding constants of conformer ensembles into a CSV:

lefqm ensembles shieldings.sdf shieldings.csv
lefqm ensembles shieldings/ shieldings.csv

Convert NMR shielding constants into chemical shifts with a linear regression
correction:

lefqm shifts shieldings.csv --calibration calibration.csv shifts.csv
"""


def get_args(args=None):
    """Get commandline arguments"""
    parser = argparse.ArgumentParser(
        prog="lefqm",
        description=DESCRIPTION,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    subparsers = parser.add_subparsers(dest="action")
    subparsers.required = True

    add_conformers_subparser(subparsers)
    add_shieldings_subparser(subparsers)
    add_ensembles_subparser(subparsers)
    add_shifts_subparser(subparsers)

    return parser.parse_args(args=args)


def call_subtool(args):
    """Call a subtool based on the args"""
    if args.verbose:
        logging.basicConfig(level=logging.INFO)
    else:
        logging.basicConfig(level=logging.WARNING)

    if args.action == "conformers":
        conformers(args)
    elif args.action == "ensembles":
        ensembles(args)
    elif args.action == "shieldings":
        shieldings(args)
    elif args.action == "shifts":
        shifts(args)
    else:
        raise RuntimeError("Invalid action")


def main():
    """Fluorine chemical shift prediction"""
    args = get_args()
    call_subtool(args)


if __name__ == "__main__":
    main()



"""Constants"""
from pathlib import Path

# import and re-export all lefshift constants
from lefshift.constants import *  # noqa

SHIELDING_COLUMN = "Shielding"
HARTREE_TO_KCALMOL = 627.5094740631
GEOMETRY_PRECISION = 1e-07

ENERGY_SD_PROPERTY = "energy in water (kcal/mol)"
BOLTZMANN_WEIGHT_SD_PROPERTY = "boltzmann weight in water"
SHIELDING_SD_PROPERTY = "isotropic shielding in water"

DEFAULT_CONFIG = Path(__file__).absolute().parent / "config.ini"



"""Omega interface functions"""
import logging
import shutil
import subprocess
import tempfile
from pathlib import Path

from rdkit import Chem
from rdkit.Chem import AllChem, SDMolSupplier, SDWriter


def omega_generate(
    mol,
    omega="omega2",
    cores=1,
    max_confs=250,
    prune_threshold=0.15,
    run_dir_path=None,
):
    """Generate conformations with omega

    Omega commandline interface:
        Simple parameter list
            Execute Options
              -param : A parameter file

            File Options
              -in : Input filename
              -out : Output filename
              -prefix : Prefix to use to name output files
              -progress : Method of showing job progress. "none","dots","log","percent".
              -sdEnergy : Writes conformer energies to the SD tag field
              -verbose : Triggers copious logging output

            3D Construction Parameters
              -fromCT : Generate structures from connection-table only.

            Torsion Driving Parameters
              -ewindow : Energy window used for conformer selection [Defaults: dense:
                         15.0]
              -maxconfs : Maximum number of conformations to be saved. [Defaults: rocs:
                          50, fastrocs: 10, dense: 20000]
              -rms : RMS threshold used to determine duplicate conformations [Defaults:
                     dense: 0.3]
              -useGPU : Switch on GPU-accelerated torsion driving (Default: true on
                        supported Linux systems with GPU)

            Stereo Parameters
              -strictstereo : Requires that all chiral atoms and bonds have specified
                              stereo [Defaults: dense: false]

            General
              -strict : A convenience flag to set -strictstereo, -strictatomtyping, and
                        -strictfrags to true or false and override previous settings.

    :param mol: mol to generate conformations for
    :type mol: rdkit.Chem.rdchem.Mol
    :param omega: path/call to conformator
    :type omega: str
    :param cores: number of cores to run on
    :type cores: int
    :param max_confs: maximum number of conformations to generate
    :type max_confs: int
    :param prune_threshold: RMSD threshold of molecule to consider duplicates
    :type prune_threshold: float
    :param run_dir_path: path to the directory to run in
    :type run_dir_path: pathlib.Path
    :return: mol with conformations
    :rtype: rdkit.Chem.rdchem.Mol
    """
    if not shutil.which(omega):
        raise RuntimeError(f"Cannot find {omega}")

    tmp_dir = None
    if run_dir_path is None:
        tmp_dir = tempfile.TemporaryDirectory()
        run_dir_path = Path(tmp_dir.name)

    logging.debug("Generating conformers in: %s", run_dir_path)

    mol = Chem.AddHs(mol)
    AllChem.EmbedMolecule(mol, clearConfs=True, randomSeed=63, useRandomCoords=True)

    input_file_path = run_dir_path / "input.sdf"
    writer = SDWriter(str(input_file_path))
    writer.SetForceV3000(True)
    writer.write(mol)
    writer.close()

    output_file_path = run_dir_path / "output.sdf"
    logging.debug("Output file is: %s", output_file_path)

    args = [
        omega,
        "-in",
        str(input_file_path),
        "-out",
        str(output_file_path),
        "-maxconfs",
        str(max_confs),
        "-rms",
        str(prune_threshold),
    ]
    if cores > 1:
        args.extend(["-mpi_np", str(cores)])

    logging.debug(" ".join(args))
    shell_output = subprocess.check_output(args, stderr=subprocess.STDOUT, cwd=run_dir_path)
    if shell_output:
        logging.debug(shell_output)

    if not output_file_path.exists():
        raise RuntimeError("Did not generate conformations")

    mols = list(SDMolSupplier(str(output_file_path), removeHs=False))
    mol = mols[0]
    for conformer_mol in mols[1:]:
        mol.AddConformer(conformer_mol.GetConformer(0), assignId=True)

    logging.info("Generated %s conformations", mol.GetNumConformers())

    if tmp_dir is not None:
        tmp_dir.cleanup()
    return mol



"""Commandline calculations"""
import tempfile
from pathlib import Path

from rdkit import Chem
from rdkit.Chem import AllChem

from lefqm.conformator import conformator_generate
from lefqm.gaussian import gaussian_calculate_shieldings
from lefqm.nwchem import nwchem_calculate_shieldings
from lefqm.omega import omega_generate
from lefqm.turbomole import turbomole_calculate_shieldings


class CommandlineCalculation:
    """Commandline calculation"""

    def __init__(self, config, run_dir_path=None):
        """Commandline calculation

        :param config: config for the calculation
        :type config: configparser.ConfigParser
        :param run_dir_path: path to the directory to run in
        :type run_dir_path: pathlib.Path
        """
        self.config = config
        self.run_dir_path = run_dir_path
        self.tmp_dir = None
        if self.run_dir_path is None:
            self.tmp_dir = tempfile.TemporaryDirectory()
            self.run_dir_path = Path(self.tmp_dir.name)

    def __del__(self):
        """Delete commandline calculation

        Remove the temporary directory if one was created.
        """
        if self.tmp_dir is not None:
            self.tmp_dir.cleanup()


class ConformerGeneration(CommandlineCalculation):
    """Conformer generation"""

    def run(self, mol):
        """Run conformer generation

        :param mol: molecule to run the generation for
        :type mol: rdkit.Chem.rdchem.Mol
        :return: mol with conformations
        :rtype: rdkit.Chem.rdchem.Mol
        """
        if self.config["Workflow"]["confgen_method"] == "conformator":
            return conformator_generate(
                mol,
                conformator=self.config["Paths"]["conformator"],
                max_confs=int(self.config["Parameters"]["max_confs"]),
                run_dir_path=self.run_dir_path,
            )
        if self.config["Workflow"]["confgen_method"] == "rdkit":
            mol = Chem.AddHs(mol)
            AllChem.EmbedMultipleConfs(mol, numConfs=int(self.config["Parameters"]["max_confs"]))
            return mol
        if self.config["Workflow"]["confgen_method"] == "omega":
            return omega_generate(
                mol,
                omega=self.config["Paths"]["omega"],
                max_confs=int(self.config["Parameters"]["max_confs"]),
                cores=int(self.config["Parameters"]["cores"])
                if self.config["Parameters"]["cores"] is not None
                else None,
                prune_threshold=float(self.config["Parameters"]["conf_prune_threshold"]),
            )
        raise RuntimeError("Invalid QM method")


class ShieldingCalculation(CommandlineCalculation):
    """NMR shielding constant calculation"""

    def run(self, mol):
        """Run shielding calculation

        :param mol: molecule to run the calcualtion on
        :type mol: rdkit.Chem.rdchem.Mol
        :return: shieldings for every atom
        :rtype: list[float]
        """
        if self.config["Workflow"]["qm_method"] == "turbomole":
            return turbomole_calculate_shieldings(
                mol,
                x2t=self.config["Paths"]["x2t"],
                ridft=self.config["Paths"]["ridft"],
                mpshift=self.config["Paths"]["mpshift"],
                cores=self.config["Parameters"]["cores"],
                run_dir_path=self.run_dir_path,
            )
        if self.config["Workflow"]["qm_method"] == "nwchem":
            return nwchem_calculate_shieldings(
                mol,
                nwchem=self.config["Paths"]["nwchem"],
                run_dir_path=self.run_dir_path,
            )
        if self.config["Workflow"]["qm_method"] == "gaussian":
            return gaussian_calculate_shieldings(
                mol,
                gaussian=self.config["Paths"]["gaussian"],
                run_dir_path=self.run_dir_path,
            )
        raise RuntimeError("Invalid QM method")
